@@ -28,12 +28,6 @@ class SharedBackendConf(SharedBackendConfUtils):
     def backendClusterMode(self):
         return self._jconf.backendClusterMode()
 
-    def cloudName(self):
-        return self._get_option(self._jconf.cloudName())
-
-    def isH2OReplEnabled(self):
-        return self._jconf.isH2OReplEnabled()
-
     def scalaIntDefaultNum(self):
         return self._jconf.scalaIntDefaultNum()
 
@@ -124,15 +118,21 @@ class SharedBackendConf(SharedBackendConfUtils):
     def flowExtraHttpHeaders(self):
         return self._get_option(self._jconf.flowExtraHttpHeaders())
 
+    def flowProxyRequestMaxSize(self):
+        return self._jconf.flowProxyRequestMaxSize()
+
+    def flowDir(self):
+        return self._get_option(self._jconf.flowDir())
+
+    def flowProxyResponseMaxSize(self):
+        return self._jconf.flowProxyResponseMaxSize()
+
     def isInternalSecureConnectionsEnabled(self):
         return self._jconf.isInternalSecureConnectionsEnabled()
 
     def isInsecureXGBoostAllowed(self):
         return self._jconf.isInsecureXGBoostAllowed()
 
-    def flowDir(self):
-        return self._get_option(self._jconf.flowDir())
-
     def clientIp(self):
         return self._get_option(self._jconf.clientIp())
 
@@ -187,6 +187,12 @@ class SharedBackendConf(SharedBackendConfUtils):
     def restApiTimeout(self):
         return self._jconf.restApiTimeout()
 
+    def cloudName(self):
+        return self._get_option(self._jconf.cloudName())
+
+    def isH2OReplEnabled(self):
+        return self._jconf.isH2OReplEnabled()
+
     def nthreads(self):
         return self._jconf.nthreads()
 
@@ -351,14 +357,14 @@ class SharedBackendConf(SharedBackendConfUtils):
         self._jconf.setMaxParallelScalaCellJobs(value)
         return self
 
-    def setInternalPortOffset(self, value):
-        self._jconf.setInternalPortOffset(value)
-        return self
-
     def setBasePort(self, value):
         self._jconf.setBasePort(value)
         return self
 
+    def setInternalPortOffset(self, value):
+        self._jconf.setInternalPortOffset(value)
+        return self
+
     def setMojoDestroyTimeout(self, value):
         self._jconf.setMojoDestroyTimeout(value)
         return self
@@ -371,6 +377,14 @@ class SharedBackendConf(SharedBackendConfUtils):
         self._jconf.setFlowExtraHttpHeaders(value)
         return self
 
+    def setFlowProxyRequestMaxSize(self, value):
+        self._jconf.setFlowProxyRequestMaxSize(value)
+        return self
+
+    def setFlowProxyResponseMaxSize(self, value):
+        self._jconf.setFlowProxyResponseMaxSize(value)
+        return self
+
     def setInternalSecureConnectionsEnabled(self):
         self._jconf.setInternalSecureConnectionsEnabled()
         return self
